/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Simple Timer + Clocks.
 *
 * The Initial Developer of the Original Code is
 *   George Bradt.
 *
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   George Bradt.
 *   Tsprajna.
 *
 * Credits:
 *   clockBlue16.png, clockRed16.png - Mark James at http://www.famfamfam.com/lab/icons/silk/
 *   clock32.png, clock48.png - David Vignoni at http://www.icon-king.com
 *   textbox style (adapted) - Regis Caspar at http://caspar.regis.free.fr/timespinner/
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var SimpleClocks = {
    prefs: null,
    popupNotify: false,
    holdPopup: 3,
    audioNotify: false,
    dialogNotify: false,
    repeatAudio: 0,
    repeatMax: 1,
    audioFile: 0,
    userWavPath: "",
    onStartup: 0,
    eventLogging: false,
    eventLogPath: "",
    deleteLogEntries: 0,
    leftClick: 0,
    disablePauseResume: false,
    tooltip: 3,
    showSecs: false,
    clock24Hr: false,
    dateFormat: 0,
    countdownSeconds: 0,
    countdownMinutes: 0,
    countdownHours: 0,
    countdownMilli: 0,
    countdownRecurring: false,
    countdowns: "",
    notifications: "",
    locations: "",
    prevTootipHeight: -1,
    availScreenHeight: 0,
    numCountdownsForTooltip: 0,
    numNotificationsForTooltip: 0,
    numLocationsForTooltip: 0,
    locationTable: [],
    countdownTable: [],
    countdownTimer: {timeDisplay: "",
                     recurring: false,
                     description: "",
                     timeMilli: 0,
                     timeFinishedMilli: 0,
                     paused: false
                    },
    notifyTable: [],
    notification: {completed: false,
                   recurring: false,
                   timeDisplay: "",
                   description: "",
                   time24hr: "",
                   timeMilli: 0,
                   timeAdjustedForDst: false,
                   currDay: -1,
                   day: -1,
                   url: "",
                   location: "",
                   sortValue: 0
                  },
    currNotifyIndex: 0,
    completedIndices: [],
    currNotifyInterval: 0,
    paused: false,
    tooltipOverrun: false,
    timerTooltipID: 0,
    timerID: 0,
    notifyID: 0,
    repeatAudioId: 0,
    repeatMaxId: 0,
    startTime: null,
    startBrowserTime: null,
    elapsedSecs: 0,
    elapsedMins: 0,
    elapsedHours: 0,
    elapsedMilli: 0,
    remainingSecs: 0,
    remainingMins: 0,
    remainingHours: 0,
    remainingMilli: 0,
    clockInProgress: false,
    countdownInProgress: false,
    countupInProgress: false,
    savedCountupTime: "",

    // Initialize the extension.

    startup: function() {
        var strbundle = document.getElementById("simclk-strings");
        try {
            // Register to receive notifications of preference changes.
            this.prefs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService).
                    getBranch("extensions.simpletimerClocks@grbradt.org.");
            this.prefs.QueryInterface(Components.interfaces.nsIPrefBranch2);
            this.prefs.addObserver("", this, false);

            this.popupNotify =         this.prefs.getBoolPref("popupNotify");
            this.holdPopup =           this.prefs.getIntPref("holdPopup");
            this.audioNotify =         this.prefs.getBoolPref("audioNotify");
            this.dialogNotify =        this.prefs.getBoolPref("dialogNotify");
            this.repeatAudio =         this.prefs.getIntPref("repeatAudio");
            this.repeatMax =           this.prefs.getIntPref("repeatMax");
            this.audioFile =           this.prefs.getIntPref("audioFile");
            this.userWavPath =         this.prefs.getCharPref("userWavPath");
            this.onStartup =           this.prefs.getIntPref("onStartup");
            this.eventLogging =        this.prefs.getBoolPref("eventLogging");
            this.eventLogPath =        this.prefs.getCharPref("eventLogPath");
            this.deleteLogEntries =    this.prefs.getIntPref("deleteLogEntries");
            this.leftClick =           this.prefs.getIntPref("leftClick");
            this.disablePauseResume =  this.prefs.getBoolPref("disablePauseResume");
            this.tooltip =             this.prefs.getIntPref("tooltip");
            this.showSecs =            this.prefs.getBoolPref("showSecs");
            this.clock24Hr =           this.prefs.getBoolPref("clock24Hr");
            this.dateFormat =          this.prefs.getIntPref("dateFormat");
            this.countdownHours =      this.prefs.getIntPref("countdownHours");
            this.countdownMinutes =    this.prefs.getIntPref("countdownMinutes");
            this.countdownSeconds =    this.prefs.getIntPref("countdownSeconds");

            // May contain Unicode characters.
            this.countdowns = this.prefs.getComplexValue("countdowns",
                    Components.interfaces.nsISupportsString).data;
            this.notifications = this.prefs.getComplexValue("notifications",
                    Components.interfaces.nsISupportsString).data;
            this.locations =   this.prefs.getComplexValue("locations",
                    Components.interfaces.nsISupportsString).data;
        }
        catch(e) {
            alert(strbundle.getString("alert.error.get.prefs") +
                  "\n" + e.message);
        }

        // Time the browser session.
        this.startBrowserTime = new Date();

        // Create default logging file.
        if ( !this.eventLogPath ) {
            SimpleClocksEventLog.createDefaultLogFile();
        }

        // Build a notifications table.
        if ( this.notifications ) {
            SimpleClocksNotify.buildNotifyTableFromPref(this.notifications);
            this.getNotifyTable();
        }

        // Build a locations table.
        // Do this before the update processing.
        SimpleClocksLocations.buildLocationsTable();

        // Will generate the "install-event-fired" pref, which is automatically
        // set to true on first run. The pref observer routine for this pref takes over from there.
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication).
                extensions.get("simpletimerClocks@grbradt.org");

        // Get the tz version file and update the "tzVersion" pref.
        // If we have a new version, the pref observer routine for this pref takes over.
        SimpleClocksLocations.readTzVersionFile();

        // Set the statusbar icon.
        this.setIconImage();

        // If there are notifications, start timing.
        if ( this.notifications ) {
            this.setNotifyInterval();
        }

        // Build a countdowns table.
        if ( this.countdowns ) {
            // Incomplete timers from previous session.
            SimpleClocksCountdown.buildCountdownTableFromPref(this.countdowns);
            this.getCountdownTable();

            if ( this.countdownTable.length > 0 ) {
                // Continue timers. Supercedes auto functions.
                this.setCountdownTimer();
                return;
            }
            else {
                // Timers from previous session have all expired. Update countdowns pref.
                this.setCountdownTable();
                SimpleClocksCountdown.updateCountdownPref();
            }
        }

        // Check if countup in progress in other browser windows.
        if ( this.countupInOtherBrowserWindow() ) {
            // Do likewise. Supercedes auto functions.
            return;
        }

        // Check for auto functions.
        switch ( this.onStartup ) {
            case 0:
                break;

            case 1:
                this.clock();
                break;

            case 2:
                this.countUp();
                break;

            default:
                alert(strbundle.getString("alert.error.invalid.startup"));
        }
    },

    // Clean up after ourselves.

    shutdown: function() {
        this.prefs.removeObserver("", this);
    },

    // Called when events occur on the preferences.

    observe: function(subject, topic, data) {
        var strbundle = document.getElementById("simclk-strings");

        if ( topic === "nsPref:changed" ) {
            switch ( data ) {
                case "popupNotify":
                    this.popupNotify = this.prefs.getBoolPref("popupNotify");
                    break;

                case "holdPopup":
                    this.holdPopup = this.prefs.getIntPref("holdPopup");
                    break;

                case "audioNotify":
                    this.audioNotify = this.prefs.getBoolPref("audioNotify");
                    break;

                case "dialogNotify":
                    this.dialogNotify = this.prefs.getBoolPref("dialogNotify");
                    break;

                case "repeatAudio":
                    this.repeatAudio = this.prefs.getIntPref("repeatAudio");
                    break;

                case "repeatMax":
                    this.repeatMax = this.prefs.getIntPref("repeatMax");
                    break;

                case "audioFile":
                    this.audioFile = this.prefs.getIntPref("audioFile");
                    break;

                case "userWavPath":
                    this.userWavPath = this.prefs.getCharPref("userWavPath");
                    break;

                case "onStartup":
                    this.onStartup = this.prefs.getIntPref("onStartup");
                    break;

                case "eventLogging":
                    this.eventLogging = this.prefs.getBoolPref("eventLogging");
                    break;

                case "eventLogPath":
                    this.eventLogPath = this.prefs.getCharPref("eventLogPath");
                    break;

                case "deleteLogEntries":
                    this.deleteLogEntries = this.prefs.getIntPref("deleteLogEntries");
                    break;

                case "leftClick":
                    this.leftClick = this.prefs.getIntPref("leftClick");
                    break;

                case "disablePauseResume":
                    this.disablePauseResume = this.prefs.getBoolPref("disablePauseResume");
                    break;

                case "tooltip":
                    this.tooltip = this.prefs.getIntPref("tooltip");
                    break;

                case "showSecs":
                    this.showSecs = this.prefs.getBoolPref("showSecs");
                    break;

                case "clock24Hr":
                    this.clock24Hr = this.prefs.getBoolPref("clock24Hr");
                    break;

                case "dateFormat":
                    this.dateFormat = this.prefs.getIntPref("dateFormat");
                    break;

                case "countdownHours":
                    this.countdownHours = this.prefs.getIntPref("countdownHours");
                    break;

                case "countdownMinutes":
                    this.countdownMinutes = this.prefs.getIntPref("countdownMinutes");
                    break;

                case "countdownSeconds":
                    this.countdownSeconds = this.prefs.getIntPref("countdownSeconds");
                    break;

                case "notifyClosed":
                    // User clicked OK.
                    if ( this.getNotifyTable() ) {
                        this.clearNotifyInterval();
                        this.setNotifyInterval();
                    }

                    break;

                case "countdowns":
                    this.countdowns = this.prefs.getComplexValue("countdowns",
                           Components.interfaces.nsISupportsString).data;
                    break;

                case "notifications":
                    this.notifications = this.prefs.getComplexValue("notifications",
                           Components.interfaces.nsISupportsString).data;
                    break;

                case "insertBefore":
                    break;

                case "locations":
                    this.locations = this.prefs.getComplexValue("locations",
                           Components.interfaces.nsISupportsString).data;
                    break;

                case "xmlRetrieved":
                    // XMLHttpRequest completed.
                    SimpleClocksLocations.updateLocationsPref();
                    SimpleClocksLocations.buildLocationsTable();

                    if ( this.notifications ) {
                        // Reset the notification time for the nots.
                        // This will force the times to be recalculated using
                        // potentially new location data.
                        this.clearNotifyInterval();
                        SimpleClocksNotify.resetNotifyPref(this.notifications);
                        SimpleClocksNotify.buildNotifyTableFromPref(this.notifications);
                        this.getNotifyTable();
                        this.setNotifyInterval();
                    }

                    break;

                case "tzVersion":
                    // New tz data file; when it's loaded, change in
                    // xmlRetrieved pref triggers locations pref update processing.
                    if ( this.locations ) {
                        this.debug("Simple Timer + Clocks: updating tz data");
                        SimpleClocksLocations.getXmlDoc();
                        var params = { displayItems: null, msg: strbundle.getString("msg.tzVersion.updated" )};
                        setTimeout(SimpleClocks.displaySliderAlert, 3000, params);
                    }

                    break;

                // Generated automatically by Application.extensions.
                // The pref will be set to false after the extension has been installed.
                case "install-event-fired":
                    this.debug("Simple Timer + Clocks: install event fired");
                    SimpleClocksLocations.readTzVersionFile();
                    break;

                default:
                    alert(strbundle.getString("alert.error.unknown.pref"));
            }
        }
    },

    // Displays a time value on statusbar panel.

    displayTime: function(hours, mins, secs) {
        var displayTime = ( hours > 0 ) ?
                           hours + ":" + this.padTime(mins) + ":" + this.padTime(secs) :
                           this.padTime(mins) + ":" + this.padTime(secs);

        document.getElementById("simclk-statpanelLabel").value = displayTime;
    },

    // For statpanel tooltips that include countdown timers, refresh every second.

    setTooltip: function() {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                        getService(Components.interfaces.fuelIApplication);

        // An array of selectedLocation objects (see SimpleClocksLocations).
        this.locationTable = Application.storage.get("locationsTable", "ERR");

        if ( this.locationTable === "ERR" ) {
            this.debug(strbundle.getString("alert.error.loading.locations"));
            return;
        }

        this.tooltipOverrun = false;
        this.prevTootipHeight = -1;
        this.availScreenHeight = window.screen.availHeight;
        this.numCountdownsForTooltip = this.countdownTable.length;
        this.numNotificationsForTooltip = this.notifyTable.length;
        this.numLocationsForTooltip = this.locationTable.length;

        this.fillTooltip();

        if ( ( this.tooltip === 2 ||
             ( this.tooltip === 4 && this.locationTable.length > 0 ) ||
             ( this.tooltip === 5 && this.countdownTable.length > 0 ) ) &&
               this.timerTooltipID === 0 ) {
            this.timerTooltipID = setInterval("SimpleClocks.fillTooltip()", 1000);
        }
    },

    // Called on statpanel mouseout.

    clearTooltip: function() {
        clearInterval(this.timerTooltipID);
        this.timerTooltipID = 0;
    },

    // Fill the statpanel tooltip based on user pref.

    fillTooltip: function() {
        var strbundle = document.getElementById("simclk-strings");

        switch ( this.tooltip ) {
            case 0:
                // No tooltip.
                if ( document.getElementById("simclk-statpanelTimer").
                			hasAttribute("tooltip") ) {
                    document.getElementById("simclk-statpanelTimer").
                    		removeAttribute("tooltip");
                }
                break;

            case 1:
                // Date.
                document.getElementById("simclk-descTtipCurrDate0").value =
                		new Date().toLocaleDateString();
                document.getElementById("simclk-statpanelTimer").
                		setAttribute("tooltip", "simclk-ttipDate");
                break;

            case 2:
                // Full tooltip.
                // Date.
                document.getElementById("simclk-descTtipCurrDate1").value =
                		new Date().toLocaleDateString();

                // Notifications.
                var parent = document.getElementById("simclk-gridNotify");
                var oldRowsNode = parent.lastChild;
                var newRowsNode = document.createElement("rows");
                var descrHeader = document.createElement("description");
                var rowNode, descrNodeTime, descrNodeDescription, descrNodeLocation;
                var msg;

                // timeDiffHrs used in not. table tooltip.
                var locationTime = {time: "", timeDiffHrs: 0};
                var prevSortValue = -1;

                if ( this.notifyID ) {
                     // Display a header line.
                    // Display "No notifications..." header for case where all nots. are deleted
                    // via deletion of locations, in which case this.notifyID still active.
                    descrHeader.setAttribute("value",
                            strbundle.getString("msg.no.notifications"));

                    for ( var g in this.notifyTable ) {
                        if ( this.notifyTable[g].completed === false ) {
                            // Active not., so override header.
                            // Give rows element an id for styling.
                            newRowsNode.setAttribute("id", "simclk-rowsNotify");

                            descrHeader.setAttribute("value",
                                    strbundle.getString("msg.notifications"));
                            break;
                        }
                    }

                    newRowsNode.appendChild(descrHeader);

                    var showHeaders = false;

                    for ( var h in this.notifyTable ) {
                        if ( this.notifyTable[h].completed === false &&
                             this.notifyTable[h].sortValue > 0 ) {
                            showHeaders = true;
                            break;
                        }
                    }

                    for ( var i = 0; i < this.numNotificationsForTooltip; i++ ) {
                        // Bypass completed nots.
                        this.notification = this.notifyTable[i];

                        if ( this.notification.completed ) {
                            continue;
                        }

                        // Check if location still being tracked.
                        if ( this.notification.location !== strbundle.getString("msg.local") &&
                             SimpleClocksLocations.searchLocationsTable(this.notification.location) === null ) {
                            this.notifyTable[i].completed = true;
                            continue;
                        }

                        if ( this.notification.sortValue !== prevSortValue &&
                             showHeaders ) {
                            // Display a location header line.
                            descrHeader = document.createElement("description");
                            descrHeader.setAttribute("value", this.notification.location);

                            rowNode = document.createElement("row");
                            rowNode.appendChild(descrHeader);
                            newRowsNode.appendChild(rowNode);
                        }

                        prevSortValue = this.notification.sortValue;

                        rowNode = document.createElement("row");

                        locationTime.time = showHeaders ? "  " + this.notification.timeDisplay :
                                this.notification.timeDisplay;
                        descrNodeTime = document.createElement("description");
                        descrNodeTime.setAttribute("value", locationTime.time);
                        rowNode.appendChild(descrNodeTime);

                        msg = this.notification.description ?
                                '"' + this.notification.description + '"' :
                                strbundle.getString("msg.no.description");
                        msg += this.notification.url ?
                                " " + "[URL]" : "";
                        msg += this.notification.day >= 0 ?
                                " / " + SimpleClocksLocations.getDow(this.notification.day) : "";
                        descrNodeDescription = document.createElement("description");
                        descrNodeDescription.setAttribute("value", msg);
                        rowNode.appendChild(descrNodeDescription);

                        newRowsNode.appendChild(rowNode);
                    }

                    if ( this.tooltipOverrun &&
                         this.numNotificationsForTooltip !== this.notifyTable.length ) {
                        rowNode = document.createElement("row");

                        descrNodeDescription = document.createElement("description");
                        descrNodeDescription.setAttribute("value", strbundle.getString("msg.etc"));
                        rowNode.appendChild(descrNodeDescription);

                        newRowsNode.appendChild(rowNode);
                    }
                }
                else {
                    // No notifications.
                    rowNode = document.createElement("row");
                    descrHeader.setAttribute("value",
                            strbundle.getString("msg.no.notifications"));

                    rowNode.appendChild(descrHeader);
                    newRowsNode.appendChild(rowNode);
                }

                parent.replaceChild(newRowsNode, oldRowsNode);

                // Clocks.
                parent = document.getElementById("simclk-gridClocks");
                oldRowsNode = parent.lastChild;
                newRowsNode = document.createElement("rows");

                descrHeader = document.createElement("description");

                if ( this.locationTable.length === 0 ) {
                    descrHeader.setAttribute("value",
                            strbundle.getString("msg.no.clocks"));
                    newRowsNode.appendChild(descrHeader);
                }
                else {
                    // Give rows element an id for styling.
                    newRowsNode.setAttribute("id", "simclk-rowsClocks");

                    descrHeader.setAttribute("value",
                            strbundle.getString("msg.clocks"));
                    newRowsNode.appendChild(descrHeader);

                    for ( var j = 0; j < this.numLocationsForTooltip; j++ ) {
                        rowNode = document.createElement("row");

                        descrNodeLocation = document.createElement("description");
                        descrNodeLocation.setAttribute("value", this.locationTable[j].location);
                        rowNode.appendChild(descrNodeLocation);

                        locationTime = SimpleClocksLocations.getLocationTime(this.locationTable[j], this.clock24Hr);
                        descrNodeTime = document.createElement("description");
                        descrNodeTime.setAttribute("value", locationTime.time);
                        rowNode.appendChild(descrNodeTime);
                        newRowsNode.appendChild(rowNode);
                    }

                    if ( this.tooltipOverrun &&
                         this.numLocationsForTooltip !== this.locationTable.length ) {
                        rowNode = document.createElement("row");

                        descrNodeLocation = document.createElement("description");
                        descrNodeLocation.setAttribute("value", strbundle.getString("msg.etc"));
                        rowNode.appendChild(descrNodeLocation);

                        newRowsNode.appendChild(rowNode);
                    }
                }

                parent.replaceChild(newRowsNode, oldRowsNode);

                // Countdown timers.
                parent = document.getElementById("simclk-gridCountdown");

                this.createTimerTooltip(parent);

                // Count up.
                msg = this.savedCountupTime ?
                      strbundle.getString("msg.last.countup") +
                      " " + this.savedCountupTime :
                      strbundle.getString("msg.no.countup");
                document.getElementById("simclk-descTtipLastCountUp").value = msg;

                // Browser time.
                document.getElementById("simclk-descTtipSessionTime").value =
                        strbundle.getString("msg.browser.time") +
                        " " + this.calcBrowserTime();

                // Test tooltip height.
                var ttip = document.getElementById("simclk-ttipFull");
                var height = parseInt(document.defaultView.getComputedStyle(ttip, null).getPropertyValue("height"), 10);

                if ( !isNaN(height) &&
                     height > (this.availScreenHeight * 0.97) ) {
                    var notifyGrid = document.getElementById("simclk-gridNotify");
                    var clocksGrid = document.getElementById("simclk-gridClocks");
                    var countdownGrid = document.getElementById("simclk-gridCountdown");
                    var notifyHeight = parseInt(document.defaultView.getComputedStyle(notifyGrid, null).getPropertyValue("height"), 10);
                    var clocksHeight = parseInt(document.defaultView.getComputedStyle(clocksGrid, null).getPropertyValue("height"), 10);
                    var countdownHeight = parseInt(document.defaultView.getComputedStyle(countdownGrid, null).getPropertyValue("height"), 10);
                    var maxHeight = Math.max(notifyHeight, clocksHeight, countdownHeight);

                    if ( maxHeight === notifyHeight ) {
                        this.numNotificationsForTooltip = Math.ceil(this.numNotificationsForTooltip / 2 );
                    }
                    else if ( maxHeight === clocksHeight ) {
                        this.numLocationsForTooltip = Math.ceil(this.numLocationsForTooltip / 2 );
                    }
                    else {
                        this.numCountdownsForTooltip = Math.ceil(this.numCountdownsForTooltip / 2 );
                    }

                    this.tooltipOverrun = true;
                }

                // Set tooltip.
                document.getElementById("simclk-statpanelTimer").
                        setAttribute("tooltip", "simclk-ttipFull");

                break;

            case 3:
                // Date, but only if clock running.
                if ( this.clockInProgress ) {
                    document.getElementById("simclk-descTtipCurrDate0").value =
                    		new Date().toLocaleDateString();
                    document.getElementById("simclk-statpanelTimer").
                    		setAttribute("tooltip", "simclk-ttipDate");
                }
                else if ( document.getElementById("simclk-statpanelTimer").
                		hasAttribute("tooltip") ) {
                    document.getElementById("simclk-statpanelTimer").
                    	removeAttribute("tooltip");
                }
                break;

            case 4:
                // Clocks, but only with clock.
                if ( !this.clockInProgress ) {
                    if ( document.getElementById("simclk-statpanelTimer").
                		     hasAttribute("tooltip") ) {
                        document.getElementById("simclk-statpanelTimer").
                    	    removeAttribute("tooltip");
                    }

                    return;
                }

                // Clocks. Create description elements, add them
                // as children to a row element (in a rows elememt in a grid), then replace old rows with new one.
                parent = document.getElementById("simclk-gridTtipClocks");
                oldRowsNode = parent.lastChild;
                newRowsNode = document.createElement("rows");
                var descrNoLocations;

                if ( this.locationTable.length === 0 ) {
                    rowNode = document.createElement("row");
                    descrNoLocations = document.createElement("description");
                    descrNoLocations.setAttribute("value",
                            strbundle.getString("msg.no.clocks"));
                    rowNode.appendChild(descrNoLocations);
                    newRowsNode.appendChild(rowNode);
                }
                else {
                    for ( var k = 0; k < this.numLocationsForTooltip; k++ ) {
                        locationTime = SimpleClocksLocations.getLocationTime(this.locationTable[k], this.clock24Hr);
                        rowNode = document.createElement("row");
                        descrNodeLocation = document.createElement("description");
                        descrNodeLocation.setAttribute("value", this.locationTable[k].location);
                        rowNode.appendChild(descrNodeLocation);
                        descrNodeTime = document.createElement("description");
                        descrNodeTime.setAttribute("value", locationTime.time);
                        rowNode.appendChild(descrNodeTime);
                        newRowsNode.appendChild(rowNode);
                    }

                    if ( this.tooltipOverrun &&
                         this.numLocationsForTooltip !== this.locationTable.length ) {
                        rowNode = document.createElement("row");

                        descrNodeLocation = document.createElement("description");
                        descrNodeLocation.setAttribute("value", strbundle.getString("msg.etc"));
                        rowNode.appendChild(descrNodeLocation);

                        newRowsNode.appendChild(rowNode);
                    }

                    // Test tooltip height.
                    ttip = document.getElementById("simclk-ttipClocks");
                    height = parseInt(document.defaultView.getComputedStyle(ttip, null).getPropertyValue("height"), 10);

                    if ( !isNaN(height) &&
                         height > (this.availScreenHeight * 0.97) &&
                         height !== this.prevTootipHeight ) {
                        this.prevTootipHeight = height;
                        this.numLocationsForTooltip = Math.ceil(this.numLocationsForTooltip / 2 );
                        this.tooltipOverrun = true;
                    }
                }

                parent.replaceChild(newRowsNode, oldRowsNode);

                // Set tooltip.
                document.getElementById("simclk-statpanelTimer").
                        setAttribute("tooltip", "simclk-ttipClocks");
                break;

            case 5:
                // Countdown timers.
                parent = document.getElementById("simclk-gridTtipCountdown");

                this.createTimerTooltip(parent);

                // Test tooltip height.
                ttip = document.getElementById("simclk-ttipCountdown");
                height = parseInt(document.defaultView.getComputedStyle(ttip, null).getPropertyValue("height"), 10);

                if ( !isNaN(height) &&
                     height > (this.availScreenHeight * 0.97) &&
                     height !== this.prevTootipHeight ) {
                    this.prevTootipHeight = height;
                    this.numCountdownsForTooltip = Math.ceil(this.numCountdownsForTooltip / 2 );
                    this.tooltipOverrun = true;
                }

                // Set tooltip.
                document.getElementById("simclk-statpanelTimer").
                        setAttribute("tooltip", "simclk-ttipCountdown");
                break;

            default:
                this.debug(strbundle.getString("alert.error.invalid.tooltip"));
        }
    },

    // Countdown timer portion of statpanel tooltip.

    createTimerTooltip: function(parent) {
        // Countdown timers. Create description elements, add them
        // as children to a row element (in a rows elememt in a grid),
        // then replace old rows with new one.
        var strbundle = document.getElementById("simclk-strings");
        var oldRowsNode = parent.lastChild;
        var newRowsNode = document.createElement("rows");
        var rowNode, descrNodeTime, descrNodeDescription;
        var msg, displayTime, countdownTimeMilli, countdownTime,
            displayHours, displayMinutes, displaySeconds;
        var descrHeader = document.createElement("description");

        if ( this.countdownTable.length === 0 ) {
            rowNode = document.createElement("row");
            descrHeader.setAttribute("value",
                    strbundle.getString("msg.no.timers"));
            rowNode.appendChild(descrHeader);
            newRowsNode.appendChild(rowNode);
        }
        else {
            if ( this.tooltip === 2 ) {
                // Give rows element an id for styling.
                newRowsNode.setAttribute("id", "simclk-rowsTimers");
            }

            // Display a header line.
            descrHeader.setAttribute("value",
                    strbundle.getString("msg.countdown.timers"));
            newRowsNode.appendChild(descrHeader);

            var currDate = new Date();
            var currTimeMilli = currDate.getTime() - currDate.getMilliseconds();

            for ( var i = 0; i < this.numCountdownsForTooltip; i++ ) {
                // Potential problem if timers expire (shrinking countdownTable size) while
                // this tooltip open, creating bounds error with this.countdownTable[i].
                try {
                    countdownTimeMilli = this.countdownTable[i].timeFinishedMilli - currTimeMilli;
                }
                catch(e) {
                    break;
                }

                rowNode = document.createElement("row");
                descrNodeTime = document.createElement("description");

                if ( i === 0 && this.paused ) {
                    displayTime = document.getElementById("simclk-statpanelLabel").value;
                    descrNodeTime.setAttribute("value", displayTime);
                }
                else {
                    countdownTime = new Date(countdownTimeMilli);
                    displayHours = countdownTime.getUTCHours();
                    displayMinutes = this.padTime(countdownTime.getUTCMinutes());
                    displaySeconds = this.padTime(countdownTime.getUTCSeconds());

                    if ( countdownTimeMilli > 0 ) {
                        displayTime = ( displayHours > 0 ) ?
                                displayHours + ":" + displayMinutes + ":" + displaySeconds :
                                displayMinutes + ":" + displaySeconds;
                    descrNodeTime.setAttribute("value", displayTime);
                    }
                    else {
                        descrNodeTime.setAttribute("value", "00:00:00");
                    }
                }

                rowNode.appendChild(descrNodeTime);

                descrNodeDescription = document.createElement("description");
                msg = this.countdownTable[i].description ?
                        '"' + this.countdownTable[i].description + '"' :
                        strbundle.getString("msg.no.description");

                msg += this.countdownTable[i].recurring ?
                        " / " + "recurring" : "";
                descrNodeDescription.setAttribute("value", msg);
                rowNode.appendChild(descrNodeDescription);

                newRowsNode.appendChild(rowNode);
            }

            if ( this.tooltipOverrun &&
                 this.numCountdownsForTooltip !== this.countdownTable.length ) {
                rowNode = document.createElement("row");

                descrNodeDescription = document.createElement("description");
                descrNodeDescription.setAttribute("value", strbundle.getString("msg.etc"));
                rowNode.appendChild(descrNodeDescription);

                newRowsNode.appendChild(rowNode);
            }
        }

        parent.replaceChild(newRowsNode, oldRowsNode);
    },

    // Check if there is a timer already in progress.
    // If last action was count up, save time for tooltip.

    cancelTimerInProgress: function() {
        var strbundle = document.getElementById("simclk-strings");

        if ( this.timerID ) {
            window.clearInterval(window.SimpleClocks.timerID);
            this.timerID = 0;
            this.startTime = null;
            document.getElementById("simclk-statpanelLabel").value = "";
        }

        if ( this.countupInProgress ) {
            this.countupInProgress = false;
            this.savedCountupTime = this.padTime(this.elapsedHours) + ":" +
                    this.padTime(this.elapsedMins) + ":" +
                    this.padTime(this.elapsedSecs);

            if ( this.eventLogging ) {
                // Pass event type, event time, recurring, status, description(N/A), URL(N/A).
                SimpleClocksEventLog.logEvent(strbundle.getString("msg.countup"),
                                              this.savedCountupTime,
                                              strbundle.getString("msg.not.applicable"),
                                              strbundle.getString("msg.completed"),
                                              strbundle.getString("msg.not.applicable"),
                                              strbundle.getString("msg.not.applicable"));
            }
        }

        this.countdownInProgress = false;
        this.clockInProgress = false;
        this.paused = false;
    },

    // Change the statusbar panel appearance.

    changeStatpanelMode: function(mode) {
        if ( mode === "wake" ) {
            // Wake up panel if sleeping.
            if ( document.getElementById("simclk-statpanelLabel").collapsed ) {
                document.getElementById("simclk-statpanelLabel").value = "";
                document.getElementById("simclk-statpanelLabel").collapsed = false;
                document.getElementById("simclk-statpanelImage").collapsed = true;
            }
        }
        else {
            // Sleep panel if awake.
            if ( document.getElementById("simclk-statpanelImage").collapsed ) {
                document.getElementById("simclk-statpanelImage").collapsed = false;
                document.getElementById("simclk-statpanelLabel").collapsed = true;
            }
        }
    },

    // Calculate elapsed time this browser session, displayed in tooltip.

    calcBrowserTime: function() {
        // Round elapsed milli to nearest second
        var browseMilli = new Date().getTime() - this.startBrowserTime.getTime();
        browseMilli = Math.round(browseMilli / 1000) * 1000;
        var browseSecs = ( browseMilli / 1000 ) % 60;
        var browseMins = (Math.floor(browseMilli / ( 1000 * 60 ))) % 60;
        var browseHours = Math.floor(browseMilli / ( 1000 * 60 * 60 ));

        return ( this.padTime(browseHours) + ":" +
                 this.padTime(browseMins) + ":" +
                 this.padTime(browseSecs) );
    },

    // Calculate elapsed time. Used by both countup and countdown.

    calcElapsedTime: function() {
        // Round elapsed milli to nearest second.
        this.elapsedMilli = new Date().getTime() - this.startTime.getTime();
        this.elapsedMilli = Math.round(this.elapsedMilli / 1000) * 1000;
        this.elapsedSecs = ( this.elapsedMilli / 1000 ) % 60;
        this.elapsedMins = (Math.floor(this.elapsedMilli / ( 1000 * 60 ))) % 60;
        this.elapsedHours = Math.floor(this.elapsedMilli / ( 1000 * 60 * 60 ));
    },

    // Calculate remaining time for countdown.

    calcRemainingTime: function() {
        this.remainingMilli  = this.countdownMilli - this.elapsedMilli;
        this.remainingSecs  = ( this.remainingMilli / 1000 ) % 60;
        this.remainingMins =  (Math.floor(this.remainingMilli / ( 1000 * 60 ))) % 60;
        this.remainingHours = Math.floor(this.remainingMilli / ( 1000 * 60 * 60 ));
    },

    // Pad a display time with zero if necessary.

    padTime: function(time) {
        return ( ( time < 10 ) ? "0" + time : time );
    },

    // Convert a clock (hh:mm:ss) to milliseconds.

    convertClockToMilli: function(clock) {
        return ( ( parseInt(clock, 10) * 60 * 60 * 1000 ) +
                 ( parseInt(clock.substr(clock.indexOf(":") + 1), 10) * 60 * 1000 ) +
                 ( parseInt(clock.substr(clock.lastIndexOf(":") + 1), 10) * 1000 ) );
    },

    // Left-click on icon is user-defined. Left-click during count up
    // or count down is pause/resume (can be disabled in Options). Middle-click is stop/reset.

    onPanelClick: function(event) {
        if ( event.button === 0 ) {
            // Process left mouse clicks.
            if ( this.countdownInProgress || this.countupInProgress ) {
                if ( !this.disablePauseResume ) {
                    this.pauseResume();
                }
            }
            else if ( this.clockInProgress ) {
                // something someday.
            }
            else {
                // Icon must be displayed.
                switch ( this.leftClick ) {
                    case 0:
                        break;

                    case 1:
                        this.clock();
                        break;

                    case 2:
                        this.countUp();
                        break;

                    case 3:
                        this.countDown();
                        break;

                    case 4:
                        this.notifyMe();
                        break;

                    default:
                        var strbundle = document.getElementById("simclk-strings");
                        alert(strbundle.getString("alert.error.invalid.leftClick"));
                }
            }
        }
        else if ( event.button === 1 &&
                ( this.timerID || this.paused )) {
            // Middle-click. Paused countup has no timer id.
            this.stop();
        }
    },

    // Called when user selects Clock menuitem, or left-click
    // statbar icon if defined.

    clock: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        while ( enumerator.hasMoreElements() ) {
            var win = enumerator.getNext();

            win.SimpleClocks.cancelTimerInProgress();

            if ( win.SimpleClocks.countdowns ) {
                win.SimpleClocks.countdownTable = [];
                win.SimpleClocks.setCountdownTable();
                win.SimpleClocksCountdown.updateCountdownPref();
            }

            win.SimpleClocks.changeStatpanelMode("wake");
            win.SimpleClocks.startTime = new Date();
            win.SimpleClocks.clockInProgress = true;

            win.SimpleClocks.timerID = win.setInterval(win.SimpleClocks.displayClock, 1000, "clock");
        }
    },

    // Display a clock on statusbar using current time (mode = "clock").
    // Return time to Timer Completed dialog (mode = "dialog").

    displayClock: function(mode) {
        var displayTime = "";
        var date = new Date();

        if ( SimpleClocks.clock24Hr ) {
            // User preference is to use 24 hr clock.
            if ( SimpleClocks.showSecs ) {
                // User preference is to show seconds.
                displayTime = SimpleClocks.padTime(date.getHours()) +
                    ":" + SimpleClocks.padTime(date.getMinutes()) +
                    ":" + SimpleClocks.padTime(date.getSeconds());
            }
            else {
                displayTime = SimpleClocks.padTime(date.getHours()) +
                    ":" + SimpleClocks.padTime(date.getMinutes());
            }
        }
        else {
            if ( SimpleClocks.showSecs ) {
                displayTime = date.toLocaleTimeString();
            }
            else {
                var time = date.toLocaleTimeString().split(" ");

                if ( time.length === 1 || time[0].indexOf(":") >= 0 ) {
                    // Local time is 24 hr clock, or 12 hr clock where
                    // time value precedes AM/PM chars (typical).
                    // Discard the seconds.
                    displayTime = time[0].substr(0, time[0].lastIndexOf(":"));

                    // If local time is 12 hr, time[1] will be AM/PM chars.
                    displayTime = ( time.length > 1 ) ?
                                displayTime + " " + time[1] : displayTime;
                }
                else {
                    // AM/PM chars. precede time value eg. Korean
                    displayTime = time[0] + " " +
                            time[1].substr(0, time[1].lastIndexOf(":"));
                }
            }
        }

        if ( mode === "clock") {
            // Append a date as indicated by user in Options.
            displayTime += SimpleClocks.getFormattedDate(date);
            document.getElementById("simclk-statpanelLabel").value = displayTime;
        }

        // Return value used in Timer Completed load routine.
        // Ignored in clock mode.
        return displayTime;
    },

    // Return a formatted date as indicated by user in Options.

    getFormattedDate: function(date) {
        switch ( this.dateFormat ) {
            case 0:
                // No date.
                return "";

            case 1:
                // eg. Fri.
                return date.toLocaleFormat(" %a");

            case 2:
                // eg. Fri, May 29.
                return date.toLocaleFormat(" %a, %b %d");

            case 3:
                // eg. Fri, May 29, 2009.
                return date.toLocaleFormat(" %a, %b %d, %Y");

            case 4:
                // eg. Fri, 05/29/09.
                return date.toLocaleFormat(" %a, %m/%d/%y");

            case 5:
                // eg. Fri, 05/29/2009.
                return date.toLocaleFormat(" %a, %m/%d/%Y");

            case 6:
                // eg. Fri, Fri, 29/05/09.
                return date.toLocaleFormat(" %a, %d/%m/%y");

            case 7:
                // eg. Fri, 29/05/2009.
                return date.toLocaleFormat(" %a, %d/%m/%Y");

            default:
                var strbundle = document.getElementById("simclk-strings");
                alert(strbundle.getString("alert.error.invalid.dateFormat"));
                return "";
        }
    },

    // Called when user selects Notify Me At menuitem.

    notifyMe: function() {
        this.setNotifyTable();

        // Display Notify Me At dialog, non-modal.
        window.openDialog(
                 "chrome://simpletimerClocks/content/simpletimerClocks-notify.xul",
                 "",
                 "centerscreen, chrome, resizable, dependent");
    },

    // Called after Notify Me At dialog closed.

    getNotifyTable: function() {
        var strbundle = document.getElementById("simclk-strings");
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        this.notifyTable = Application.storage.get("notifyTable", "ERR");

        if ( this.notifyTable === "ERR" ) {
            alert(strbundle.getString("alert.error.loading.notifications"));
            return false;
        }
        else {
            return true;
        }
    },

    // Called before Notify Me At dialog opened.

    setNotifyTable: function() {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        Application.storage.set("notifyTable", this.notifyTable);

    },

    // Convert notification time to milli.
    // Take into account locale and dst if necessary.

    calcNotTimeMilli: function(notTime24hr, location, currNotTimeMilli,
                               isTimeAdjustedForDst, notTime) {
        var strbundle = document.getElementById("simclk-strings");

        var notMilli;
        var notTimeHrsMins = notTime24hr.split(":");
        var currDate = new Date();
        var year = currDate.getUTCFullYear();
        var month = currDate.getUTCMonth();
        var day = currDate.getUTCDate();
        var hours = currDate.getUTCHours();
        var mins = currDate.getUTCMinutes();

        var currMilli = Date.UTC(year, month, day, hours, mins);

        notTime.dstInEffect = false;

        if ( location !== strbundle.getString("msg.local") ) {
            var trackedLocation = SimpleClocksLocations.searchLocationsTable(location);

            if ( trackedLocation === null ) {
                // User likely no longer tracking this location.
                return -1;
            }

            if ( currNotTimeMilli === 0 ) {
                // First time this not. processed.
                // Note the use of a radix in parseInt, otherwise
                // leading 0 values get treated as octal. (eg "08" "09" return 0)
                var offset = SimpleClocksLocations.convertMilliToClock(trackedLocation.offsetMilli).split(":");
                var notHours = parseInt(notTimeHrsMins[0], 10) - parseInt(offset[0], 10);
                var notMins;

                if ( parseInt(offset[0], 10) < 0 ) {
                    notMins = parseInt(notTimeHrsMins[1], 10) + parseInt(offset[1], 10);
                }
                else {
                    notMins = parseInt(notTimeHrsMins[1], 10) - parseInt(offset[1], 10);
                }


                notMilli = Date.UTC(year, month, day, notHours, notMins);
            }
            else {
                notMilli = currNotTimeMilli;
            }

            notMilli = this.adjustNotMilli(currMilli, notMilli);

            if ( trackedLocation.observesDst ) {
                // Dst in effect.
                if ( SimpleClocksLocations.currentlyDst(trackedLocation, notMilli) ) {
                    notTime.dstInEffect = true;

                    if ( !isTimeAdjustedForDst ) {
                        // First time processed, or dst just started.
                        notMilli = notMilli + trackedLocation.offsetMilli -
                                trackedLocation.dstOffsetMilli;
                        notMilli = this.adjustNotMilli(currMilli, notMilli);
                    }
                }

                if ( isTimeAdjustedForDst && !notTime.dstInEffect ) {
                    // dst just ended.
                    notMilli = notMilli - trackedLocation.offsetMilli +
                            trackedLocation.dstOffsetMilli;
                }
            }

            // Find the day of not. in the tracked location,
            // which may be different from local not. day.
            var offClock;

            if ( notTime.dstInEffect ) {
                offClock = SimpleClocksLocations.convertMilliToClock(trackedLocation.dstOffsetMilli * -1);
            }
            else {
                offClock = SimpleClocksLocations.convertMilliToClock(trackedLocation.offsetMilli * -1);
            }

            var tempDate = new Date(notMilli).toUTCString();
            notTime.day = new Date(tempDate + " " + offClock).getUTCDay();
        }
        else {
            // Local.
            var notDate;

            if ( currNotTimeMilli === 0 ) {
                // First time this not. processed.
                notDate = new Date(currDate.getFullYear(), currDate.getMonth(),
                        currDate.getDate(), parseInt(notTimeHrsMins[0], 10), parseInt(notTimeHrsMins[1], 10));
            }
            else {
                notDate = new Date(currNotTimeMilli);
            }

            while ( currDate.getTime() > notDate.getTime() ){
                // Also adjusts for dst if necessary.
                notDate.setDate(notDate.getDate() + 1);
            }

            notMilli = notDate.getTime();
            notTime.day = notDate.getDay();
        }

        return notMilli;
    },

    // After calculating notMilli, make sure it is within the next 24 hours.

    adjustNotMilli: function(currMilli, notMilli) {
        var dayMilli = 24 * 60 * 60 * 1000;

        while ( currMilli >= notMilli ) {
            // Add a day.
            notMilli += dayMilli;
        }

        while ( ( notMilli - currMilli ) > dayMilli ) {
            // Subtract a day.
            notMilli -= dayMilli;
        }

        return notMilli;
    },

    // Set icon image based on pending nots.

    setIconImage: function() {
        // Check if any pending nots., and set icon image
        var status = "normal";
        var len = this.notifyTable.length;

        for ( var i = 0; i < len; i++ ) {
            if ( this.notifyTable[i].completed === false ) {
                status = "notify";
                break;
            }
        }

        document.getElementById("simclk-statpanelImage").
        		setAttribute("status", status);
    },

    // Clear notify interval.

    clearNotifyInterval: function() {
        this.setIconImage();
        clearInterval(this.notifyID);
        this.notifyID = 0;
    },

    // Set notify interval based on soonest pending notify.

    setNotifyInterval: function() {
        var strbundle = document.getElementById("simclk-strings");

        if ( !this.notifyTable || this.notifyTable.length === 0 ) {
            return;
        }

        var intervalTime = 0;

        // For the not. selected to time, this is the index into the not. table.
        this.currNotifyIndex = 0;

        // For the not. selected to time, this is the time until completion.
        this.currNotifyInterval = 0;

        var calcTimeMilli;

        // These properties are passed to and updated in this.calcNotTimeMilli.
        var notTime = {day: -1, dstInEffect: false};

        var currDate = new Date();
        var currTimeMilli = currDate.getTime() -
                ( currDate.getSeconds() * 1000 ) - currDate.getMilliseconds();

        // Find the not. with the smallest interval time.
        for ( var i in this.notifyTable ) {
            // Bypass completed nots.
            if ( this.notifyTable[i].completed === true ) {
                continue;
            }

            calcTimeMilli = this.calcNotTimeMilli(this.notifyTable[i].time24hr,
                                                  this.notifyTable[i].location,
                                                  parseInt(this.notifyTable[i].timeMilli, 10),
                                                  this.notifyTable[i].timeAdjustedForDst,
                                                  notTime);
            intervalTime = calcTimeMilli - currTimeMilli;

            if ( calcTimeMilli === -1 ) {
                // Notification for locale which is no longer tracked.
                this.notifyTable[i].completed = true;
                continue;
            }
            else if ( calcTimeMilli <= 0 ||
                     intervalTime <= 0 ) {
                // Oops.
                alert(strbundle.getString("alert.error.notification.time"));

                this.debug("Simple Timer + Clocks");
                this.debug("Current time: " + currDate.getTime());
                this.debug("Current millisec time (not including secs): " + currTimeMilli);
                this.debug("Calculated notification time (milli): " + calcTimeMilli);
                this.debug("Time until notification (milli): " + intervalTime);
                this.dumpNotificationObject(this.notifyTable[i]);

                continue;
            }

            if ( this.currNotifyInterval === 0 ||
                 intervalTime < this.currNotifyInterval ) {
                this.currNotifyInterval = intervalTime;
                this.currNotifyIndex = i;
            }


           this.notifyTable[i].timeAdjustedForDst = notTime.dstInEffect;
           this.notifyTable[i].currDay = notTime.day;
           this.notifyTable[i].timeMilli = calcTimeMilli;
        }

        // No point checking for completion every second if we are more than
        // a minute from completion.
        if ( this.currNotifyInterval ) {
            this.notifyID = ( this.currNotifyInterval > ( 60 * 1000 ) ) ?
                    setTimeout("SimpleClocks.switchNotifyInterval()",
                    		(this.currNotifyInterval - ( 60 * 1000 ))) :
                    setInterval("SimpleClocks.monitorLastMinuteNotify()", 1000);
        }

        // Update notifications pref.
        this.setNotifyTable();
        SimpleClocksNotify.updateNotifyPref();
    },

    // Change intervals when its one minute prior to notify. Called once.

    switchNotifyInterval: function() {
        clearTimeout(this.notifyID);
        this.notifyID = setInterval("SimpleClocks.monitorLastMinuteNotify()", 1000);
    },

    // Is it time to notify the user?

    monitorLastMinuteNotify: function() {
        var currDate = new Date();

        if ( currDate.getTime() > this.notifyTable[this.currNotifyIndex].timeMilli ) {
            // We're done.
            var timeoutIncrement = 0;
            this.completedIndices = [];

            this.clearNotifyInterval();

            // Check if multiple nots. expired at same time.
            // Add a 3 second delay between the alerts.
            for ( var i in this.notifyTable ) {
                if ( this.notifyTable[i].completed === false &&
                     currDate.getTime() > this.notifyTable[i].timeMilli ) {
                    this.completedIndices.push(i);
                    setTimeout("SimpleClocks.processCompletedNotify()", timeoutIncrement);
                    setTimeout("SimpleClocks.setIconImage()", timeoutIncrement + 1000);

                    if ( !this.notifyTable[i].recurring || this.notifyToday(i) ) {
                        timeoutIncrement +=  3000;
                    }
                }
            }

            timeoutIncrement += 1000;
            setTimeout("SimpleClocks.setNotifyInterval()", timeoutIncrement);
        }
    },

    // A completed notification is not deleted from table,
    // but updated.

    processCompletedNotify: function() {
        var strbundle = document.getElementById("simclk-strings");

        if ( this.completedIndices.length > 0 ) {
            // Multiple nots. expired at the same time.
            // Grab (and remove) the first element (not. table index).
            this.currNotifyIndex = this.completedIndices.shift();
        }

        // Check if location still being tracked.
        var location = this.notifyTable[this.currNotifyIndex].location;

        if ( location !== strbundle.getString("msg.local") &&
             SimpleClocksLocations.searchLocationsTable(location) === null ) {
            this.notifyTable[this.currNotifyIndex].completed = true;
            return;
        }

        var recurring = this.notifyTable[this.currNotifyIndex].recurring;

        if ( !recurring || this.notifyToday(this.currNotifyIndex) ) {
            this.checkUserPrefs("notify");

            // Link if URL provided but dialog alarm not used.
            if ( this.notifyTable[this.currNotifyIndex].url  &&
                 !this.dialogNotify &&
                 ( this.audioNotify ||
                   this.popupNotify ) ) {
                this.autoOpenLink(this.notifyTable[this.currNotifyIndex].url);
            }

            if ( this.eventLogging ) {
                var time = this.notifyTable[this.currNotifyIndex].location + " " +
                        this.notifyTable[this.currNotifyIndex].timeDisplay;
                var logRecurring = ( recurring ) ?
                        strbundle.getString("msg.yes") :
		    			strbundle.getString("msg.no");
                var description = ( this.notifyTable[this.currNotifyIndex].description ) ?
                        this.notifyTable[this.currNotifyIndex].description :
                        strbundle.getString("msg.none");
                var url = ( this.notifyTable[this.currNotifyIndex].url ) ?
                        this.notifyTable[this.currNotifyIndex].url :
                        strbundle.getString("msg.none");

                // Pass event type, event time, recurring, status, description, URL.
                SimpleClocksEventLog.logEvent(strbundle.getString("msg.notification"),
                                              time,
                                              logRecurring,
                                              strbundle.getString("msg.completed"),
                                              description,
                                              url);
            }
        }

        if ( recurring ) {
            // Recurring ones don't get marked completed.
            if ( location === strbundle.getString("msg.local") ) {
                // Add 1 day to not. milli time. This accounts for local dst
                // so if the new time is the first day of dst, added 25 hours.
                var notDate = new Date(this.notifyTable[this.currNotifyIndex].timeMilli);
                notDate.setDate(notDate.getDate() + 1);
                this.notifyTable[this.currNotifyIndex].timeMilli = notDate.getTime();
            }
            else {
                // Add 24 hours to not. milli time. We don't want to account for local dst
                // when dealing with other time zones.
                this.notifyTable[this.currNotifyIndex].timeMilli +=  1000 * 60 * 60 * 24;
            }
        }
        else {
            // Mark this notification as completed.
            this.notifyTable[this.currNotifyIndex].completed = true;
        }

        // Update notifications pref.
        this.setNotifyTable();
        SimpleClocksNotify.updateNotifyPref();
    },

    // Is today a notification day?

    notifyToday: function(index) {
        // Not. day values:
        // 0-6 for Sun-Sat,
        // 7 daily, 8 weekdays, 9 weekends.
        var currDay = this.notifyTable[index].currDay;
        var notDay = this.notifyTable[index].day;

        if ( notDay === 7 || notDay === currDay ||
             ( notDay === 8 && ( currDay > 0 && currDay < 6 ) ) ||
             ( notDay === 9 && ( currDay === 0 || currDay === 6 ) ) ) {
            return true;
        }
        else {
            return false;
        }
    },

    // Called when user selects Count Down menuitem.

    countDown: function() {
        this.setCountdownTable();

        // Display time entry dialog, non-modal.
        window.openDialog(
                 "chrome://simpletimerClocks/content/simpletimerClocks-countdown.xul",
                 "",
                 "centerscreen, chrome, resizable, dependent");
    },

    // Called after Countdown time entry dialog closed (via OK).

    getCountdownTable: function() {
        var strbundle = document.getElementById("simclk-strings");
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        var table = Application.storage.get("countdownTable", "ERR");

        if ( table === "ERR" ) {
            alert(strbundle.getString("alert.error.loading.countdowns"));
            return false;
        }
        else {
            // Independent copy of the array of countdown objects.
            this.countdownTable = this.deepCopy(table);
            return true;
        }
    },

    // Called before Countdown time entry dialog opened.

    setCountdownTable: function() {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        Application.storage.set("countdownTable", this.countdownTable);

    },

    // Start countdown.

    setCountdownTimer: function() {
        if ( this.countdownTable.length === 0 ) {
            SimpleClocks.cancelTimerInProgress();
            SimpleClocks.changeStatpanelMode("sleep");

            // Update countdowns pref.
            this.setCountdownTable();
            SimpleClocksCountdown.updateCountdownPref();

            return;
        }

        this.startTime = new Date();
        var currTimeMilli = this.startTime.getTime() - this.startTime.getMilliseconds();

        for ( var i in this.countdownTable ) {
            // For new timers set the completion time.
            // Paused timer will stay on top (timeFinishedMilli = 0) until resumed.
            if ( this.countdownTable[i].timeFinishedMilli === 0 &&
                 !this.countdownTable[i].paused ) {
                this.countdownTable[i].timeFinishedMilli =
                        currTimeMilli + this.countdownTable[i].timeMilli;
            }
        }

        // Sort the table. If the first entry is recurring, it may no longer be next to expire.
        if ( this.countdownTable.length > 1 ) {
            this.countdownTable.sort(SimpleClocksCountdown.sortCountdownTable);
        }

        // Update countdowns pref.
        this.setCountdownTable();
        SimpleClocksCountdown.updateCountdownPref();

        if ( this.countdownTable[0].paused ) {
            // timeMilli was set to remainingMilli when paused.
            this.paused = true;
            this.countdownMilli = this.countdownTable[0].timeMilli;
        }
        else {
            this.paused = false;
            this.countdownMilli = this.countdownTable[0].timeFinishedMilli - currTimeMilli;
        }

        if ( this.countdownMilli < 0 ) {
            // Sanity.
            this.countdownMilli = 0;
        }

        var time = new Date(this.countdownMilli);
        this.countdownHours = time.getUTCHours();
        this.countdownMinutes = time.getUTCMinutes();
        this.countdownSeconds = time.getUTCSeconds();

        SimpleClocks.changeStatpanelMode("wake");
        this.displayTime(this.countdownHours, this.countdownMinutes,
                         this.countdownSeconds);

        this.countdownInProgress = true;

        // Perform decreaseTimer() every second.
        this.timerID = setInterval("SimpleClocks.decreaseTimer()", 1000);
    },

    // Display the updated timer.

    decreaseTimer: function() {
        var strbundle = document.getElementById("simclk-strings");

        if ( !this.paused ) {
            this.calcElapsedTime();
        }

        if ( this.elapsedMilli < this.countdownMilli ) {
            // Check if others (besides first timer) have expired.
            // Possible if timer paused.
            if ( this.countdownTable.length > 1 ) {
                var currDate = new Date();
                var currTimeMilli = currDate.getTime() - currDate.getMilliseconds();
                var len = this.countdownTable.length;

                for ( var i = 1; i < len; i++ ) {
                    if ( currTimeMilli >= this.countdownTable[i].timeFinishedMilli ) {
                        // We're done.
                        this.checkUserPrefs(i); // Pass index.

                        if ( this.eventLogging ) {
                            var recurring = ( this.countdownTable[i].recurring ) ?
                                    strbundle.getString("msg.yes") :
		    			            strbundle.getString("msg.no");
                            var description = ( this.countdownTable[i].description ) ?
                                    this.countdownTable[i].description :
                                    strbundle.getString("msg.none");

                            // Pass event type, event time, recurring, status, description, URL(N/A).
                            SimpleClocksEventLog.logEvent(strbundle.getString("msg.countdown"),
                                                          this.countdownTable[i].timeDisplay,
                                                          recurring,
                                                          strbundle.getString("msg.completed"),
                                                          description,
                                                          strbundle.getString("msg.not.applicable"));
                        }

                        if ( this.countdownTable[i].recurring ) {
                            // Cannot increment with timeMilli, since it changes if paused.
                            this.countdownTable[i].timeFinishedMilli +=
                                    this.convertClockToMilli(this.countdownTable[i].timeDisplay);
                        }
                        else {
                            // Remove the completed timer.
                            this.countdownTable.splice(i, 1);

                            // Update countdowns pref.
                            this.setCountdownTable();
                            SimpleClocksCountdown.updateCountdownPref();
                        }
                    }
                }
            }

            if ( !this.paused ) {
                // Update the statpanel display.
                this.calcRemainingTime();
                this.displayTime(this.remainingHours, this.remainingMins,
                                 this.remainingSecs);
            }
        }
        else {
            // We're done.
            this.cancelTimerInProgress();
            this.displayTime(0, 0, 0);
            this.checkUserPrefs("0"); // Pass index.

            if ( this.eventLogging ) {
                recurring = ( this.countdownTable[0].recurring ) ?
                        strbundle.getString("msg.yes") :
		    			strbundle.getString("msg.no");
                var description = ( this.countdownTable[0].description ) ?
                        this.countdownTable[0].description :
                        strbundle.getString("msg.none");

                // Pass event type, event time, recurring, status, description, URL(N/A).
                SimpleClocksEventLog.logEvent(strbundle.getString("msg.countdown"),
                                              this.countdownTable[0].timeDisplay,
                                              recurring,
                                              strbundle.getString("msg.completed"),
                                              description,
                                              strbundle.getString("msg.not.applicable"));
            }

            if ( this.countdownTable[0].recurring ) {
                // Cannot increment with timeMilli, since it changes if paused.
                this.countdownTable[0].timeFinishedMilli +=
                        this.convertClockToMilli(this.countdownTable[0].timeDisplay);
                this.setCountdownTimer();
            }
            else {
                if ( this.countdownTable.length < 2 ) {
                    this.countdownTable = [];

                    // Update countdowns pref.
                    this.setCountdownTable();
                    SimpleClocksCountdown.updateCountdownPref();

                    this.changeStatpanelMode("sleep");
                }
                else {
                    // Remove first element, and start the next one.
                    this.countdownTable.shift();
                    this.setCountdownTimer();
                }
            }
        }
    },

    // Perform alerts if requested.

    checkUserPrefs: function(mode) {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].
                getService(Components.interfaces.nsIWindowMediator);
        var win = wm.getMostRecentWindow("navigator:browser");

        // Compare window references.
        if ( win.window == window.window ) {
            // Only perform the alerts once.
            if ( this.audioNotify ) {
                this.playSound(this.audioFile);
            }

            if ( this.popupNotify ) {
                params = { displayItems: null, msg: this.getDisplayMsg(mode) };
				SimpleClocksSliderAlert.addMessageToQueue(params);
             }

            if ( this.dialogNotify ) {
                this.displayTimerCompleteDlg(mode);
            }
        }
    },

    // Play the audio notification wav file.
    // This can be performed on countdown or notify completion,
    // or if user clicks on audio icon in options dlg.
    // Also performed when user requests audio repeated.

    playSound: function(audioFile) {
        var strbundle = document.getElementById("simclk-strings");
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService).
                getBranch("extensions.simpletimerClocks@grbradt.org.");
        var userWavPath = prefs.getCharPref("userWavPath");
        var selectedAudioFile;

        // Check which audio file user selected.
        switch ( audioFile ) {
            case 0:
                selectedAudioFile = "alarm_clock";
                break;

            case 1:
                selectedAudioFile = "fanfare";
                break;

            case 2:
                selectedAudioFile = "custom";
                break;

            case 3:
                selectedAudioFile = "notify";
                break;

            default:
                selectedAudioFile = "alarm_clock";
        }

        // Check if user clicked custom audio icon prior
        // to setting a custom wav.
        if ( selectedAudioFile === "custom" &&
             userWavPath === "" ) {
            alert(strbundle.getString("alert.error.no.custom.file"));
            return;
        }

        var soundURL = ( selectedAudioFile === "custom" ) ?
                    "file:///" + userWavPath :
                    "chrome://simpletimerClocks/content/sounds/" +
                    selectedAudioFile + ".wav";

        // The nsISound play method takes a URI as an argument,
        // so first convert the URL to URI.
        try {
            var sound = Components.classes["@mozilla.org/sound;1"].
                        createInstance(Components.interfaces.nsISound);
            var ios = Components.classes["@mozilla.org/network/io-service;1"].
                        getService(Components.interfaces.nsIIOService);
            var uri = ios.newURI(soundURL, null, null);
            sound.init();
            sound.play(uri);
        }
        catch (e) {
            alert(strbundle.getString("alert.error.audio.failed") +
                  "\n" + e.name);
        }
    },

    // Display the timer completed dialog.

    displayTimerCompleteDlg: function(mode) {
        // Set the message to display in dialog.
        if ( this.audioNotify &&
             !this.repeatAudioId &&
             this.repeatMax > 0 &&
             this.repeatAudio !== 0 ) {
            // Only one audio timer no matter how many
            // dialogs stack up.
            // repeatAudio > 0 is minutes, < 0 is seconds.
            var repeatAudioInterval = this.repeatAudio > 0 ?
                    this.repeatAudio * 60 * 1000 :
                    this.repeatAudio * -1 * 1000;
            var repeatMaxInterval = this.repeatAudio > 0 ?
                    repeatAudioInterval * this.repeatMax + ( 30 * 1000 ) :
                    repeatAudioInterval * this.repeatMax + 500;
            this.repeatAudioId =
                    setInterval("SimpleClocks.playSound(SimpleClocks.audioFile);",
                    repeatAudioInterval);
            this.repeatMaxId =
                    setTimeout("SimpleClocks.clearTimerCompleteInterval();",
                    repeatMaxInterval);
        }

        // It seems odd that the dlgs directly overlap with
        // no offset. So manually set offset.
        try {
            var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].
                    getService(Components.interfaces.nsIWindowMediator);
            var enumerator = wm.getEnumerator("dialog:timerComplete");

            while(enumerator.hasMoreElements()) {
                var win = enumerator.getNext();
                win.moveBy(-26, -26);
            }
        }
        catch (e) {
            alert(strbundle.getString("alert.error.dialog.failed") +
                  "\n" + e.name);
        }

        var msg = this.getDisplayMsg(mode);
        var url = this.getDisplayUrl(mode);
        var params = {inn: {msg: msg, url: url}, out: null};

        // Non-modal.
        window.openDialog(
                "chrome://simpletimerClocks/content/simpletimerClocks-timerComplete.xul",
                "",
                "centerscreen, chrome, resizable, dependent",
                params);
    },

    // Called when user closes Timer Completed dialog,
    // or when max repeat audio cycles reached.

    clearTimerCompleteInterval: function() {
        if ( this.repeatAudioId ) {
            clearInterval(this.repeatAudioId);
            this.repeatAudioId = 0;
        }

        if ( this.repeatMaxId ) {
            clearTimeout(this.repeatMaxId);
            this.repeatMaxId = 0;
        }
    },

    // Get the message to be displayed.

    getDisplayMsg: function(mode) {
        var strbundle = document.getElementById("simclk-strings");

        if ( mode === "notify" ) {
            // If user supplied a description, use it.
            if ( this.notifyTable[this.currNotifyIndex].location === strbundle.getString("msg.local") ) {
                return this.notifyTable[this.currNotifyIndex].description ?
                        this.notifyTable[this.currNotifyIndex].description :
                        strbundle.getString("alert.notification.notify.complete");
            }
            else {
                // Prepend locale, locale not. time.
                return this.notifyTable[this.currNotifyIndex].location + " " +
                        this.notifyTable[this.currNotifyIndex].timeDisplay + " " +
                        ( this.notifyTable[this.currNotifyIndex].description ?
                          this.notifyTable[this.currNotifyIndex].description :
                          strbundle.getString("alert.notification.notify.complete") );
            }
        }
        else {
            return this.countdownTable[mode].description ?
                   this.countdownTable[mode].description :
                   strbundle.getString("alert.notification.timer.complete");
        }
    },

    // Get the URL, if any, to be displayed.

    getDisplayUrl: function(mode) {
        if ( mode === "notify" ) {
            return this.notifyTable[this.currNotifyIndex].url ?
                    this.notifyTable[this.currNotifyIndex].url : "";
        }
        else {
            return "";
        }
    },

    // Called when user selects Count Up menuitem.

    countUp: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        while ( enumerator.hasMoreElements() ) {
            var win = enumerator.getNext();

            win.SimpleClocks.cancelTimerInProgress();

            if ( win.SimpleClocks.countdowns ) {
                win.SimpleClocks.countdownTable = [];
                win.SimpleClocks.setCountdownTable();
                win.SimpleClocksCountdown.updateCountdownPref();
            }

            win.SimpleClocks.changeStatpanelMode("wake");
            win.SimpleClocks.displayTime(0, 0, 0);
            win.SimpleClocks.startTime = new Date();
            win.SimpleClocks.countupInProgress = true;

            win.SimpleClocks.timerID = win.setInterval(win.SimpleClocks.increaseTimer, 1000);
        }
    },

    // Increment and display the timer.

    increaseTimer: function() {
        SimpleClocks.calcElapsedTime();
        SimpleClocks.displayTime(SimpleClocks.elapsedHours, SimpleClocks.elapsedMins, SimpleClocks.elapsedSecs);
    },

    // Called when user selects Pause/Resume menuitem, or via middle-click on panel.

    pauseResume: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        if ( this.countdownInProgress || this.countupInProgress ) {
            while ( enumerator.hasMoreElements() ) {
                var win = enumerator.getNext();
                if ( win.SimpleClocks.countupInProgress || win.SimpleClocks.countdownInProgress ) {
                    if ( win.SimpleClocks.paused ) {
                        win.SimpleClocks.resume();
                    }
                    else {
                        win.SimpleClocks.pause();
                    }
                }
            }
        }
    },

    // Stop the current count up/count down, leave the panel display intact.

    pause: function() {
        if ( SimpleClocks.countupInProgress ) {
            clearInterval(SimpleClocks.timerID);
            SimpleClocks.timerID = 0;
            SimpleClocks.startTime = null;
        }
        else {
            // Reset timeMilli to the remaining time. When resumed, it will be
            // processed as if a new timer (although the display time is unchanged).
            SimpleClocks.countdownTable[0].timeMilli = SimpleClocks.remainingMilli;
            SimpleClocks.countdownTable[0].timeFinishedMilli = 0;
            SimpleClocks.countdownTable[0].paused = true;

            // Update countdowns pref.
            SimpleClocks.setCountdownTable();
            SimpleClocksCountdown.updateCountdownPref();
        }

        SimpleClocks.paused = true;
    },

    // Resume the current count up/count down.

    resume: function() {
        SimpleClocks.paused = false;

        if ( SimpleClocks.countupInProgress ) {
            SimpleClocks.startTime = new Date(new Date().getTime() - SimpleClocks.elapsedMilli);
            SimpleClocks.timerID = setInterval("SimpleClocks.increaseTimer()", 1000);
        }
        else {
            this.cancelTimerInProgress();

            SimpleClocks.countdownMilli = SimpleClocks.countdownTable[0].timeMilli;
            SimpleClocks.countdownTable[0].paused = false;

            // Countdowns pref.updated there.
            SimpleClocks.setCountdownTimer();
        }
    },

    // Called when user selects Stop/Reset menuitem, or via middle-click on panel.

    stop: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        this.cancelTimerInProgress();

        while ( enumerator.hasMoreElements() ) {
            var win = enumerator.getNext();
            if ( win.SimpleClocks.countdownTable.length > 0 ) {
                win.SimpleClocks.countdownTable.shift();
                win.SimpleClocks.setCountdownTimer();
            }
            else {
                win.SimpleClocks.changeStatpanelMode("sleep");
            }
        }
    },

    // Displays the options dialog.

    displayOptionsDlg: function() {
        window.openDialog(
                "chrome://simpletimerClocks/content/simpletimerClocks-options.xul",
                "",
                "centerscreen, toolbar, titlebar, chrome, resizable, dependent");
    },

    // Displays the Event Log dialog.

    displayEventLogDlg: function() {
        window.openDialog(
                "chrome://simpletimerClocks/content/simpletimerClocks-eventLog.xul",
                "",
                "centerscreen, chrome, resizable, dependent");
    },

    // Displays the About dialog.

    displayAboutDlg: function() {
        window.openDialog(
                "chrome://simpletimerClocks/content/simpletimerClocks-about.xul",
                "",
                "centerscreen, chrome, resizable, dependent");
    },

    // Generic open page in new tab.
    // This function developed by Devon Jensen, from Download Statusbar.

    openLink: function(aPage) {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].
        		getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var win = wmed.getMostRecentWindow("navigator:browser");

        if ( !win ) {
            win = window.openDialog(
                    "chrome://browser/content/browser.xul",
                    "_blank",
                    "chrome,all,dialog=no",
                    aPage, null, null);
        }
        else {
            var content = win.document.getElementById("content");
            content.selectedTab = content.addTab(aPage);
        }
    },

    // Open page in new tab.
    // Use this to auto-open (completed notification) a page for each browser window open.

    autoOpenLink: function(aPage) {
        var content = document.getElementById("content");
        content.selectedTab = content.addTab(aPage);
    },

    // Check if other browser windows open.
    // Don't display the startup slider alerts if true.

    otherBrowserWindowOpen: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        if ( enumerator.hasMoreElements() ) {
            // Skip oldest window.
            enumerator.getNext();
        }

        return enumerator.hasMoreElements();
    },

    // If other browser windows open, check if countup in progress.
    // If so, mirror it.

    countupInOtherBrowserWindow: function() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        if ( enumerator.hasMoreElements() ) {
            // Oldest first.
            var win = enumerator.getNext();

            if ( win.SimpleClocks.countupInProgress ) {
                this.countupInProgress = true;
                this.paused =  win.SimpleClocks.paused;
                this.changeStatpanelMode("wake");

                if ( this.paused ) {
                    this.elapsedMilli = win.SimpleClocks.elapsedMilli;
                    this.elapsedHours = win.SimpleClocks.elapsedHours;
                    this.elapsedMins = win.SimpleClocks.elapsedMins;
                    this.elapsedSecs = win.SimpleClocks.elapsedSecs;
                    this.displayTime(this.elapsedHours, this.elapsedMins, this.elapsedSecs);
                }
                else {
                    this.startTime =  win.SimpleClocks.startTime;
                    this.timerID = window.setInterval(window.SimpleClocks.increaseTimer, 1000);
                }

                return true;
            }
        }

        return false;
    },

    // Deep copy an array of objects.

    deepCopy: function(inArray) {
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);
        var outArray = [];

        for ( var i in inArray ) {
            outArray.push(json.decode(json.encode(inArray[i])));
        }

        return outArray;
    },

    // Dump to console.

    dumpNotificationObject: function(notification) {
        this.debug("\nNotification data:");
        this.debug("Completed?: " + notification.completed);
        this.debug("Recurring?: " + notification.recurring);
        this.debug("Notification time: " + notification.timeDisplay);
        this.debug("Description: " + notification.description);
        this.debug("Notification time (24 hr format): " + notification.time24hr);
        this.debug("Notification time (millisecs): " + notification.timeMilli);
        this.debug("Notification time adjusted for dst?: " + notification.timeAdjustedForDst);
        this.debug("Notification day for current timeMilli: " + notification.currDay);
        this.debug("Notification day: " + notification.day);
        this.debug("Url: " + notification.url);
        this.debug("Locale: " + notification.location);
        this.debug("Sort value: " + notification.sortValue);
    },

    // Dump to console.

    dumpCountdownObject: function(countdownTimer) {
        this.debug("\nCountdown timer data:");
        this.debug("Countdown Time for Display: " + countdownTimer.timeDisplay);
        this.debug("Recurring?: " + countdownTimer.recurring);
        this.debug("Description: " + countdownTimer.description);
        this.debug("Countdown time (millisecs): " + countdownTimer.timeMilli);
        this.debug("Countdown finish time (millisecs): " + countdownTimer.timeFinishedMilli);
    },

    // Debug messages to console.

    debug: function(aMsg) {
        setTimeout(function() { throw new Error("[debug] " + aMsg); }, 0);
    }
};
